/**
  *  \file client/cscreen.cc
  *  \brief Control Screens, and Related Things
  *
  *  This contains the main classes for control screens.
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "client/chartusr.h"
#include "client/cscreen.h"
#include "client/dlg-list.h"
#include "client/playerscr.h"
#include "client/screens.h"
#include "cpluslib/format.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/clip.h"
#include "gfx/palette.h"
#include "resmgr/resmgr.h"
#include "ui/layout.h"
#include "ui/skin.h"
#include "ui/table.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets.h"
#include "util/nls.h"
#include "client/wscanchart.h"
#include "client/history.h"
#include "client/dlg-visual.h"
#include "int/intglobal.h"
#include "client/dlg-comment.h"
#include "client/dialogs/searchdlg.h"

//#define BLUESKIN
#ifndef BLUESKIN
#  define ICOLOR (COLOR_SHIELD+3)
#  define IFILL GRAY50_FILL
#else
#  define ICOLOR (COLOR_SHIELD+1)
#  define IFILL SOLID_FILL
#  define FCOLOR (COLOR_SHIELD+5)
#endif

/************************** WControlScreenSkin ***************************/

/** \class WControlScreenSkin
    \brief Skin for control screens */

const UIColorSet&
WControlScreenSkin::getColors() const
{
    return DARK_COLOR_SET;
}

#ifdef BLUESKIN
void
WControlScreenSkin::outText(GfxCanvas& can,  int f, int x, int y, string_t text)
{
    if (isSolid()) {
        int width = fonts[f]->getTextWidth(text);
        drawBackground(can, GfxRect(x - width*parm.just_x/2, y, width, font_heights[f]));
    }
    parm.bgcolor = GFX_TRANSPARENT;
    parm.font = fonts[f];
    ::outText(can, parm, x, y, text);
}

void
WControlScreenSkin::outTextF(GfxCanvas& can, int f, int x, int y, int lim, string_t text)
{
    if (isSolid())
        drawBackground(can, GfxRect(x - lim*parm.just_x/2, y, lim, font_heights[f]));
    parm.bgcolor = GFX_TRANSPARENT;
    parm.font = fonts[f];
    ::outTextF(can, parm, x, y, lim, text);
}

void
WControlScreenSkin::drawBackground(GfxCanvas& can, GfxRect r)
{
    can.drawBar(r, standard_colors[ICOLOR], standard_colors[0], IFILL, GFX_ALPHA_OPAQUE);
}
#else
void
WControlScreenSkin::outText(GfxCanvas& can,  int f, int x, int y, string_t text)
{
    parm.bgcolor = isSolid() ? standard_colors[COLOR_DARK] : GFX_TRANSPARENT;
    parm.font = fonts[f];
    ::outText(can, parm, x, y, text);
}

void
WControlScreenSkin::outTextF(GfxCanvas& can, int f, int x, int y, int lim, string_t text)
{
    parm.bgcolor = isSolid() ? standard_colors[COLOR_DARK] : GFX_TRANSPARENT;
    parm.font = fonts[f];
    ::outTextF(can, parm, x, y, lim, text);
}

void
WControlScreenSkin::drawBackground(GfxCanvas& can, GfxRect r)
{
    drawSolidBar(can, r, standard_colors[COLOR_DARK]);
}
#endif

/***************************** WControlScreen ****************************/

/*
  Layout:

    ,-----------------------------------.
    |              ,------------------. |   <-- borders are 5 pixels
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              `------------------' |  <-- tile_new_y  \ 3 lines for
    |              [                  ] |  ___ chart_y2    / scanner widget
    |                                   |
    |                                   |    8 lines free here for tiles
    |                                   |
    |                                   |
    `-----------------------------------'  ---

    |  320 pixels  |
                chart_x1
 */

/** Constructor.
    \param type object type we're displaying. Used for iteration.
    \param screen_id this screen's identifier.

    Checklist for derived constructors:
    - create all tiles and add them with addTile(). You should include a
      WControlScreenHeaderTile.
    - create starchart widget and hand it to addChart() */
WControlScreen::WControlScreen(GObjectType& type, WScreenId::ScreenId screen_id)
    : scanner(),
      screen_id(screen_id),
      pos_changed_conn(type.sig_pos_changed.add(this, &WControlScreen::onPosChanged)),
      type(type),
      chart(0),
      controls(UINullLayout::instance)
{
    setSkin(skin);
    setExtent(GfxRect(*ui_screen), true);
    setGeometry();
    setState(st_Modal, true);

    add(controls, extent);

    const int scanner_height = 2*font_heights[FONT_NORMAL];
    add(scanner, GfxRect(chart_x1+2,                chart_y2 + 5,
                         extent.w - chart_x1 - 5-2, scanner_height));

    pushHistoryScreenId(WScreenId(screen_id, type.getCurrent()));
}

WControlScreen::~WControlScreen()
{ }

/** Compute geometry parameters. */
void
WControlScreen::setGeometry()
{
    tile_width = 315;           // FIXME: should probably be related to font size somehow
    tile_new_y = extent.h - 8*font_heights[FONT_NORMAL];
    tile_y     = 5;
    tile_x     = 5;
    chart_y2   = tile_new_y - 3*font_heights[FONT_NORMAL];
    chart_x1   = tile_width + tile_x + 5;
}

/** Add a tile, and give it a title. */
void
WControlScreen::addTile(WTile& tile, const char* title)
{
    tile.setHeading(title);
    tile.setDisplayState(WTile::sNormal);
    addTile(tile);
}

/** Add a tile. */
void
WControlScreen::addTile(WTile& tile)
{
    UIWidget::LayoutInfo info;
    tile.getLayoutInfo(info);

    if (tile_y + info.pref_size.y > extent.h - 5) {
        /* tile does not fit in this column. Would it fit when we
           remove the heading? */
        if (tile.getDisplayState() == WTile::sNormal) {
            tile.setDisplayState(WTile::sContentOnly);
            tile.getLayoutInfo(info);
            if (tile_y + info.pref_size.y <= extent.h - 5)
                goto success;
            tile.setDisplayState(WTile::sNormal);
            tile.getLayoutInfo(info);
        }

        /* start new column */
        tile_y = tile_new_y;
        tile_x += tile_width;

        /* does it fit there? */
        if (tile_y + info.pref_size.y > extent.h - 5) {
            /* nope. Remove heading if there, but we cannot do
               anything else so let it be. */
            if (tile.getDisplayState() == WTile::sNormal) {
                tile.setDisplayState(WTile::sContentOnly);
                tile.getLayoutInfo(info);
            }
        }
    }
 success:
    GfxRect tile_size(tile_x, tile_y, tile_width, info.pref_size.y);
    tile.setExtent(tile_size, true);
    tile_y += info.pref_size.y + 2;

    add(tile, tile_size);
}

/** Add starchart widget. */
void
WControlScreen::addChart(WScannerChartWidget& chart)
{
    this->chart = &chart;
    add(chart, GfxRect(chart_x1, 5, extent.w - chart_x1 - 5, chart_y2 - 5));
    scanner.setChart(chart);
}

/** Go to object #id. This one is tolerant in that it silently ignores
    invalid Ids. */
void
WControlScreen::setCurrent(int id)
{
    if (id == type.getCurrent() || !type.isExistant(id))
        return;
    type.setCurrent(id);
}

GPoint
WControlScreen::getScannerPosition() const
{
    return chart->getScannerPosition();
}

void
WControlScreen::drawContent(GfxCanvas& can)
{
    // FIXME: optimize!
    getSkin().startDraw();
    getSkin().drawBackground(can, extent);

#ifdef BLUESKIN
    getSkin().getParameters().color = standard_colors[FCOLOR];
    drawRectangle(can, getSkin(), GfxRect(extent.x, extent.y+1, extent.w, extent.h-2));
    drawRectangle(can, getSkin(), GfxRect(extent.x+1, extent.y, extent.w-2, extent.h));

    drawRectangle(can, getSkin(), GfxRect(chart_x1-1, 5-2,        extent.w - chart_x1 - 3, chart_y2 - 5+4));
    drawRectangle(can, getSkin(), GfxRect(chart_x1-2, 5-1,        extent.w - chart_x1 - 1, chart_y2 - 5+2));

    drawRectangle(can, getSkin(), GfxRect(chart_x1-1, chart_y2+4, extent.w - chart_x1 - 3, 2*font_heights[FONT_NORMAL]+4));
#else
    drawFrameUp(can, extent);
    drawFrameDown(can, GfxRect(chart_x1-1, 5-1,        extent.w - chart_x1 - 3, chart_y2 - 5+2));
    drawFrameDown(can, GfxRect(chart_x1-1, chart_y2+4, extent.w - chart_x1 - 3, 2*font_heights[FONT_NORMAL]+4));
#endif
    can.update(extent);
}

bool
WControlScreen::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    return UIFixedWidget::handleCommand(cmd, sender);
}

bool
WControlScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch(e.key.code) {
         case SDLK_PAGEUP:
         case '-':
         case vk_WheelUp:
            setCurrent(type.getPreviousWrap(type.getCurrent(), false));
            return true;
         case SDLK_PAGEDOWN:
         case '+':
         case vk_WheelDown:
            setCurrent(type.getNextWrap(type.getCurrent(), false));
            return true;
         case SDLK_PAGEUP | ss_Ctrl:
         case '-' | ss_Ctrl:
            setCurrent(type.getPreviousWrap(type.getCurrent(), true));
            return true;
         case SDLK_PAGEDOWN | ss_Ctrl:
         case '+' | ss_Ctrl:
            setCurrent(type.getNextWrap(type.getCurrent(), true));
            return true;
         case SDLK_ESCAPE:
            stop(0);
            return true;
         case '.': {
            GObject& g = type.getObject(type.getCurrent());
            g.setIsMarked(!g.isMarked());
            type.getUniverse()->doScreenUpdates();
            return true;
         }
         case SDLK_F1:
            doScreen(WScreenId::ShipScreen, *this);
            return true;
         case SDLK_F1 + ss_Shift:
         case SDLK_F1 + ss_Ctrl:
            if (chart->isValid()) {
                const GPoint pos = ((e.key.code & ss_Ctrl)
                                    ? chart->getScannerPosition()
                                    : chart->getCenterPosition());
                const int start_ship = (screen_id==1 || screen_id==11) ? type.getCurrent() : 0;
                if (getDisplayedTurn().getCurrentUniverse().getNumberOfPlayedShipsAt(pos) > 0) {
                    GObjectId id = doListShips(pos,
                                               ls_EarlyExit,
                                               start_ship,
                                               _("Visual Scanner"),
                                               _("Go to"),
                                               0 /* FIXME */);
                    if (id.isValid())
                        stopWithScreenId(WScreenId(WScreenId::ShipScreen, id.getId()));
                }
            }
            return true;
         case SDLK_F2:
            doScreen(WScreenId::PlanetScreen, *this);
            return true;
         case SDLK_F2 + ss_Shift:
            stopWithScreenId(WScreenId(WScreenId::PlanetScreen,
                                       getDisplayedTurn().getCurrentUniverse().getPlanetAt(chart->getCenterPosition(), true)));
            return true;
         case SDLK_F2 + ss_Ctrl:
            stopWithScreenId(WScreenId(WScreenId::PlanetScreen,
                                       scanner.getScannedPlanetId(false)));
            return true;
         case SDLK_F3:
            doScreen(WScreenId::BaseScreen, *this);
            return true;
         case SDLK_F3 + ss_Shift:
            stopWithScreenId(WScreenId(WScreenId::BaseScreen,
                                       getDisplayedTurn().getCurrentUniverse().getPlanetAt(chart->getCenterPosition(), true)));
            return true;
         case SDLK_F3 + ss_Ctrl:
         case SDLK_F8 + ss_Ctrl:
            stopWithScreenId(WScreenId(WScreenId::BaseScreen,
                                       scanner.getScannedPlanetId(false)));
            return true;
         case SDLK_F4:
         case SDLK_F4 + ss_Shift:
            /* Go to starchart. If we don't have a correct current
               position, we go to the starchart anyway, using the last
               position in that case. */
            if (chart->isValid()) {
                setStarchartObject(type.getObject(type.getCurrent()));
                stopWithScreenId(WScreenId(chart->getCenterPosition()));
            } else {
                stopWithScreenId(WScreenId(getStarchartPosition()));
            }
            return true;
         case SDLK_F4 + ss_Ctrl:
            if (chart->isValid()) {
                stopWithScreenId(chart->getScannerPosition());
            }
            return true;
         case SDLK_F5:
         case SDLK_F5 + ss_Shift:
            if (int pid = getDisplayedTurn().getCurrentUniverse().getPlanetAt(chart->getCenterPosition()))
                stopWithScreenId(doPlanetScan(pid));
            return true;
         case SDLK_F5 + ss_Ctrl:
            if (int pid = scanner.getScannedPlanetId(false))
                stopWithScreenId(doPlanetScan(pid));
            return true;
         case SDLK_F7:
            {
                WScreenId id = doSearch();
                if (id.isValid())
                    stopWithScreenId(id);
            }
            return true;
         case ss_Ctrl + 'h':
            {
                WScreenId id = doSearchHere(chart->getCenterPosition());
                if (id.isValid())
                    stopWithScreenId(id);
            }
            return true;
         case SDLK_F9 + ss_Shift:
            if (int pid = getDisplayedTurn().getCurrentUniverse().getPlanetAt(chart->getCenterPosition()))
                doEditComment(getDisplayedTurn().getPlanet(pid),
                              getPlanetProperties(pid),
                              _("Enter new comment for this planet:"),
                              _("Edit Comment"));
            return true;
         case SDLK_F9 + ss_Ctrl:
            if (int pid = scanner.getScannedPlanetId(false))
                doEditComment(getDisplayedTurn().getPlanet(pid),
                              getPlanetProperties(pid),
                              _("Enter new comment for this planet:"),
                              _("Edit Comment"));
            return true;
         case 'l':
            if (chart->isValid()) {
                GObjectId id = doListShips(chart->getScannerPosition(),
                                           ls_IncludeForeignShips,
                                           0,
                                           _("Visual Scanner"),
                                           _("Go to"),
                                           0 /* FIXME */);
                if (id.isValid())
                    stopWithScreenId(WScreenId(WScreenId::ShipScreen, id.getId()));
            }
            return true;
         case 'L':
            if (chart->isValid()) {
                GObjectId id = doListShips(chart->getCenterPosition(),
                                           ls_IncludeForeignShips,
                                           0,
                                           _("Visual Scanner"),
                                           _("Go to"),
                                           0 /* FIXME */);
                if (id.isValid())
                    stopWithScreenId(WScreenId(WScreenId::ShipScreen, id.getId()));
            }
            return true;
         case 'y':
            if (chart->isValid())
                chart->doKeyboardMode();
            return true;
         case SDLK_BACKSPACE + ss_Alt: {
            WScreenId id = doScreenHistoryDialog(true);
            if (id.isNonNull())
                stopWithScreenId(id);
            return true;
         }
         case SDLK_BACKSPACE: {
            WScreenId id = popHistoryScreenId(true);
            if (id.isNonNull())
                stopWithScreenId(id);
            return true;
         }
        }
    }
    return UIFixedWidget::handleEvent(e, second_pass);
}

/** Run this control screen.
    \return Id of screen to activate (1..3 for ship/planet/base, 0
    for race screen, etc.) */
WScreenId
WControlScreen::runScreen()
{
    update();
    ui_root->add(*this);
    UIFixedWidget::run();
    ui_root->close(*this);
    return return_value;
}

void
WControlScreen::stopWithScreenId(WScreenId id)
{
    if (id.isValid()) {
        return_value = id;
        stop(0);
    }
}

/** Selection-change handler. This is called when the user picks a
    different object on this object's type. */
void
WControlScreen::onPosChanged(int /*id*/)
{
    pushHistoryScreenId(WScreenId(screen_id, type.getCurrent()));
    update();
}

/************************ WControlScreenHeaderTile ***********************/

/*! \class WControlScreenHeaderTile
    \brief Header of a Control Screen

    Displays the top-most part of a control screen: name, description,
    picture, buttons. */

/** Constructor.
    \param type object type to observe
    \param buttons the buttons you want to be displayed */
WControlScreenHeaderTile::WControlScreenHeaderTile(GObjectType& type, int buttons)
      // FIXME: magic numbers
    : WObjectTypeObserverTile(GfxPoint(315, 145), type), buttons(buttons)
{
    setLayout(LeftJustified, 5, false);
    addControl(GfxRect( 5-5,  45, 30, 25),
               holder.add(new UIButton("F1", SDLK_F1)));
    addControl(GfxRect(40-5,  45, 30, 25),
               holder.add(new UIButton("F2", SDLK_F2)));
    addControl(GfxRect( 5-5,  75, 30, 25),
               holder.add(new UIButton("F3", SDLK_F3)));
    addControl(GfxRect(40-5,  75, 30, 25),
               holder.add(new UIButton("F4", SDLK_F4)));
//     addControl(GfxRect( 5-5, 105, 30, 25),
//                holder.add(new UIButton("F6", SDLK_F6)));
    addControl(GfxRect(40-5, 105, 30, 25),
               holder.add(new UIButton("F7", SDLK_F7)));
    addControl(GfxRect(265-5, 114, 45, 25),
               holder.add(new UIButton("ESC", SDLK_ESCAPE)));
    addControl(GfxRect(230-5, 114, 25, 25),
               holder.add(new UIButton("H", 'h')));
    addControl(GfxRect(285-5, 45, 25, 25),
               holder.add(new UIButton(cm_CS_Xchg, "X", 'x')).addKey());
    addControl(GfxRect(75-5, 45, 20, 25),
               holder.add(new UIButton("\036", '-')));
    addControl(GfxRect(75-5, 75, 20, 25),
               holder.add(new UIButton("\037", '+')));
//     if (buttons & bSend)
//         addControl(GfxRect(75-5, 105, 20, 25),
//                    holder.add(new UIButton("I", 'i')));
//     if (buttons & bAuto)
//         addControl(GfxRect(230-5, 45, 50, 25),
//                    holder.add(new UIButton("Auto", SDLK_RETURN)));
    if (buttons & bCycle)
        addControl(GfxRect(275-5, 75, 35, 25),
                   holder.add(new UIButton("\035", SDLK_TAB)));
    if (buttons & bAdd)
        addControl(GfxRect(230-5, 75, 40, 25),
                   holder.add(new UIButton("Add", SDLK_INSERT)));
    if (buttons & bName)
        addControl(GfxRect(315-20-5, 0, 20, 20),
                   holder.add(new UIButton("N", 'n')));
}

void
WControlScreenHeaderTile::drawData(GfxCanvas& can)
{
    // FIXME: magic numbers
    int wi = (buttons & bName) ? extent.w-20 : extent.w;
    setColor(UIColor::tc_Heading);
    outTextF(can, FONT_TITLE, 0, 0, wi, getTitle());
    setColor(UIColor::tc_Yellow);
    outTextF(can, FONT_NORMAL, 0, 25, extent.w, getSubtitle());

    drawFrameDown(can, GfxRect(extent.x + 108, extent.y + 45, 107, 95));

    Ptr<GfxPixmap> pix = ResManager::getInstance().getPixmap(getPictureId());
    if (pix)
        pix->blitSized(can, extent.x + 109, extent.y + 46, 105, 93);

    drawColorFrame(can, GfxRect(extent.x + 108, extent.y + 45, 107, 95),
                   getObject()->isMarked() ? cf_Yellow : cf_None,
                   false);
}

/****************************** WHeadingTile *****************************/

/*! \class WHeadingTile
    \brief Section divider

    Displays a constant text string in big font, as a divider between
    several sections on a control screen.

    This tile is now superfluous, as every tile can display a heading. */

WHeadingTile::WHeadingTile(const string_t& s)
    : WTile(GfxPoint(315, font_heights[FONT_TITLE])), text(s)
{ }

WHeadingTile::WHeadingTile(const char* s)
    : WTile(GfxPoint(315, font_heights[FONT_TITLE])), text(s)
{ }

void
WHeadingTile::drawData(GfxCanvas& can)
{
    setColor(UIColor::tc_Heading);
    outTextF(can, FONT_TITLE, 0, 0, extent.w, text);
}

/******************************** WScanner *******************************/

/** \class WScanner
    \brief Scanner Evaluation

    This widget displays the result of a scan (distance, name of
    object pointed-to). It must, after construction, be associated
    with a scanner. */

/** Constructor. */
WScanner::WScanner()
    : chart(0),
      button_l("L", 'l'),
      button_y("Y", 'y')
{
    addChildWidget(button_l, 0);
    addChildWidget(button_y, 0);
}

WScanner::~WScanner()
{ }

/** Associate this scanner with a starchart widget.
    \param chart the starchart widget
    \pre this routine must not already have been called */
void
WScanner::setChart(WScannerChartWidget& chart)
{
    ASSERT(!this->chart);
    this->chart = &chart;
    update_conn = chart.sig_moved.add(this, &WScanner::update);
    update();
}

/** Update. This recomputes all displayed information, and requests a
    redraw of the widget if needed. */
void
WScanner::update()
{
    if (!chart || !chart->isValid()) {
        sid = xpid = gpid = 0;
        if (UIBaseWidget* w = getParent())
            w->setStateOf(cm_CS_Xchg, st_Disabled, true);
    } else {
        GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
        sid  = univ.getAnyShipAt(chart->getScannerPosition());
        xpid = univ.getPlanetAt(chart->getScannerPosition());
        if (xpid)
            gpid = xpid;
        else
            gpid = univ.getGravityPlanetAt(chart->getScannerPosition());
        if (UIBaseWidget* w = getParent())
            w->setStateOf(cm_CS_Xchg, st_Disabled, (xpid == 0 && gpid == 0 && sid == 0));
    }
    drawWidget(false);
}

void
WScanner::drawContent(GfxCanvas& can)
{
    const GfxRect& e = getExtent();
    UITableWriter w(can, getSkin(), GfxRect(e.x, e.y, e.w-47, e.h), FONT_NORMAL, UITableWriter::fl_Fill);
    string_t scan, dist;

    w.newLColumn();
    w.writeFixedCell(UIColor::tc_Static, _("Scan:"));
    w.writeFixedCell(UIColor::tc_Static, _("Distance:"));

    w.newLColumn();
    if (chart) {
        if (!chart->isValid()) {
            scan = dist = "...";
        } else {
            // FIXME: can we share this with GGameTurn::getLocationName()?
            GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
            if (xpid)
                scan = univ.getPlanet(xpid).getFullName();
            else if (sid)
                scan = univ.getShip(sid).getFullName();
            else if (gpid)
                scan = format(_("near %s"), univ.getPlanet(gpid).getFullName());
            else
                scan = format(_("Deep Space (%d,%d)"),
                              chart->getScannerPosition().x,
                              chart->getScannerPosition().y);
            dist = format(_("%.2f ly"),
                          std::sqrt(double(distanceSquaredRaw(chart->getScannerPosition(),
                                                              chart->getCenterPosition()))));
        }
    }
    w.writeVariableCell(UIColor::tc_Green, scan);
    w.writeVariableCell(UIColor::tc_Green, dist);
    w.finish();
    can.update(getExtent());
}

void
WScanner::getLayoutInfo(LayoutInfo& info)
{
    // FIXME?
    info.flags = 0;
    info.min_size = GfxPoint(240, 2*font_heights[FONT_NORMAL]);
    info.pref_size = GfxPoint(310, 2*font_heights[FONT_NORMAL]);
}

void
WScanner::onResize(GfxRect /*oldSize*/, bool /*byLayout*/)
{
    const GfxRect& e = getExtent();
    button_y.setExtent(GfxRect(e.getRightX()-22, e.y+2, 20, 20), true);
    button_l.setExtent(GfxRect(e.getRightX()-47, e.y+2, 20, 20), true);
}
