/**
  *  \file client/tiles/shipcargo.cc
  *  \brief Ship Cargo Tile
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/shipcargo.h"

#include "cpluslib/format.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "ui/util.h"            // drawFrameDown
#include "ui/widgets.h"
#include "util/misc.h"          // clansToString
#include "util/nls.h"
#include "game/cmds.h"
#include "game/config.h"
#include "game/global.h"

namespace {
    /** Ship cargo transfer status button. Displays a symbol when a cargo transfer
        is active, and generates an appropriate key event when the user clicks it. */
    class WShipCargoButton : public UIAbstractButton,
                             public GObjectTypeObserver
    {
        string_t text;
        bool ship;

        void set(UIKey key, string_t text);
     public:
        WShipCargoButton(GObjectType& type, bool ship);
        void drawContent(GfxCanvas& can);
        void getLayoutInfo(LayoutInfo& info);

        void onObjectChanged();
    };
}

/** Constructor.
    \param type Object type to observe
    \param ship true: observe ship/ship transporter. false: observe ship/planet transporter */
WShipCargoButton::WShipCargoButton(GObjectType& type, bool ship)
    : UIAbstractButton(),
      GObjectTypeObserver(type),
      text(),
      ship(ship)
{
    onObjectChanged();
}

/** Set content.
    \param key Key to generate
    \param text Text to display */
void
WShipCargoButton::set(UIKey key, string_t text)
{
    setKey(key);
    if (text != this->text) {
        this->text = text;
        drawWidget();
    }
}

void
WShipCargoButton::drawContent(GfxCanvas& can)
{
    /* Right-justified, yellow */
    UISkin& sk = getSkin();
    sk.startDraw();
    sk.getParameters().just_x = 2;
    sk.getParameters().just_y = 0;
    sk.setColor(UIColor::tc_Yellow);
    sk.setSolid(true);
    sk.outTextF(can, FONT_SMALL, getExtent().getRightX(), getExtent().y, getExtent().w, text);

    can.update(getExtent());
}

void
WShipCargoButton::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = font_sizes[FONT_SMALL].scaledBy(2, 1);
    info.flags = 0;
}

void
WShipCargoButton::onObjectChanged()
{
    GShip* s = dynamic_cast<GShip*>(getObject());
    if (!s || !s->isPlayed()) {
        set(0, string_t());
    } else if (ship) {
        /* We're observing the ship->ship transporter */
        if (s->isTransporterActive(GShip::t_Transfer)) {
            /* Ship transfer: "->S", Ctrl+S */
            set(ss_Ctrl + 's', "\032S");
        } else {
            set(0, string_t());
        }
    } else {
        /* We're observing the ship->planet transporter */
        bool bum = config.AllowBeamUpMultiple()
            && getDisplayedTurn().getCommands(s->getOwner()).getCommand(GCommand::phc_Beamup, s->getId()) != 0;

        if (s->isTransporterActive(GShip::t_Unload)) {
            if (s->getTransporter(GShip::t_Unload).id == 0) {
                /* Jettison: "->J", Ctrl+J */
                set(ss_Ctrl + 'j', "\032J");
            } else if (bum) {
                /* Beam up + down: "<->P", Ctrl+P */
                set(ss_Ctrl + 'p', "\035J");
            } else {
                /* Beam down: "->P", Ctrl+P */
                set(ss_Ctrl + 'p', "\032P");
            }
        } else {
            if (bum) {
                /* Beam-up only: "<-P", no hot-key */
                set(0, "\033P");
            } else {
                set(0, string_t());
            }
        }
    }
}

/***************************** WShipCargoTile ****************************/

WShipCargoTile::WShipCargoTile(GObjectType& type)
    // FIXME: magic numbers
    : WObjectTypeObserverTile(GfxPoint(315, 4 * font_heights[FONT_NORMAL] + 14 + 3), type)
{
    setLayout(RightJustified, 0, false);
    addControl(GfxRect(25, 0, 20, 20), h.add(new UIButton("C", 'c')));
    addControl(GfxRect(0,  0, 20, 20), h.add(new UIButton("U", 'u')));
    addControl(GfxRect(25, 25, 20, 15), h.add(new WShipCargoButton(type, false)));
    addControl(GfxRect(25, 40, 20, 15), h.add(new WShipCargoButton(type, true)));
}

void
WShipCargoTile::drawShipCargoTile(UIAbstractTableWriter& tw)
{
    // Column 1: ore names
    // Column 2: ore amounts
    // Column 3: space
    // Column 4: rest
    // Column 5: rest amounts
    // Column 6: units

    GShip* s = dynamic_cast<GShip*>(getObject());
    if (!s)
        return;

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, "Neu:");
    tw.writeFixedCell(UIColor::tc_Static, "Tri:");
    tw.writeFixedCell(UIColor::tc_Static, "Dur:");
    tw.writeFixedCell(UIColor::tc_Static, "Mol:");

    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("9999 kt"));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", s->getCargo(el_Neutronium)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", s->getCargo(el_Tritanium)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", s->getCargo(el_Duranium)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", s->getCargo(el_Molybdenum)));

    tw.newLColumn(5);           // spacer

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Colonists:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Supplies:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Money:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Total Mass:"));

    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("999,999"));
    tw.writeVariableCell(UIColor::tc_Green, clansToString(s->getCargo(el_Colonists)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(s->getCargo(el_Supplies)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(s->getCargo(el_Money)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(s->getMass()));

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Green, "");
    tw.writeFixedCell(UIColor::tc_Green, "kt");
    tw.writeFixedCell(UIColor::tc_Green, "mc");
    tw.writeFixedCell(UIColor::tc_Green, "kt");

    tw.finish();
}

void
WShipCargoTile::drawData(GfxCanvas& can)
{
    GShip* s = dynamic_cast<GShip*>(getObject());
    if (!s)
        return;

    GfxRect pos = getInnerSize(0, 0);
    const int y = font_heights[FONT_NORMAL];

    UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL, 0);
    drawShipCargoTile(tw);

    int cargo = s->getHull().getCargo();
    drawCargoBar(can, pos.x + 5,   pos.y + 4*y + 4, 130, s->getCargo(el_Neutronium), s->getHull().getFuel(), COLOR_GREENBLACK, _("Fuel: %d free"));
    drawCargoBar(can, pos.x + 145, pos.y + 4*y + 4, 130, cargo - s->getFreeCargo(), cargo, COLOR_BLUE, _("Cargo: %d free"));
}

void
WShipCargoTile::drawCargoBar(GfxCanvas& can, int x, int y, int wi, int have, int max,
                             int color, const char* fmt)
{
    drawFrameDown(can, GfxRect(x, y, wi, 10));
    ++x;
    ++y;
    wi -= 2;

    int split = max ? wi * have / max : 0;
    if (have > 0 && split == 0)
        split = 1;
    if (have >= max)
        split = wi;
    else if (split >= wi)
        --split;

    if (split)
        drawSolidBar(can, GfxRect(x, y, split, 8), standard_colors[color]);
    if (split != wi)
        drawSolidBar(can, GfxRect(x + split, y, wi-split, 8), standard_colors[COLOR_BLACK]);

    getSkin().getParameters().just_x = 1;
    getSkin().getParameters().just_y = 0;
    if (have > max)
        getSkin().getParameters().color = standard_colors[COLOR_RED];
    else
        getSkin().getParameters().color = standard_colors[COLOR_WHITE];
    getSkin().getParameters().bgcolor = GFX_TRANSPARENT;
    getSkin().getParameters().font = fonts[FONT_SMALL];
    ::outTextF(can, getSkin(), x + wi/2, y-1, wi, format(fmt, max-have));
}

