/**
  *  \file int/if/baseif.cc
  *  \brief Interpreter Game Interface: Starbase Properties
  */

#include "int/if/baseif.h"
#include "int/value.h"
#include "cpluslib/format.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/tables.h"
#include "game/ship.h"
#include "int/compound.h"
#include "int/if/ifutil.h"
#include "int/interror.h"
#include "game/config.h"

namespace {
    /** Starbase array property. This handles a starbase's array properties, Storage.Xxx(). */
    class IntBaseArrayProperty : public IntIndexableValue {
        const GPlanet& p;
        IntBaseProperty ibp;

        IntValue* performArrayReference(int (GPlanet::*func)(int) const, int32_t limit, int32_t arg);
     public:
        IntBaseArrayProperty(const GPlanet& p, IntBaseProperty ibp);

        // IntIndexableValue:
        IntValue* get(IntDataSegment& dseg, uint32_t index, uint32_t nargs);
        void set(IntDataSegment& dseg, uint32_t index, uint32_t nargs, IntValue* value);
        IntContext* makeFirstContext();

        // IntValue:
        IntBaseArrayProperty* clone();
        string_t toString(bool readable);
        void store(IntTagNode& sv, Stream& aux);
    };
}

/** Constructor.
    \param p Starbase
    \param ibp Property to implement */
IntBaseArrayProperty::IntBaseArrayProperty(const GPlanet& p, IntBaseProperty ibp)
    : p(p), ibp(ibp)
{ }

/** Perform array reference. This implements the special abilities of starbase array properties:
    - index 0 counts all items
    - index 1..limit returns that item's count
    - other values yield empty (not an error!)
    \param func GPlanet function to access the property in raw form (e.g. getBaseEngineStore)
    \param limit Maximum index (e.g. NUM_ENGINES)
    \param arg User-specified argument
    \return resulting value */
IntValue*
IntBaseArrayProperty::performArrayReference(int (GPlanet::*func)(int) const, int32_t limit, int32_t arg)
{
    if (arg == 0) {
        int sum = 0;
        for (int i = 1; i <= limit; ++i)
            sum += (p.*func)(i);
        return makeIntValue(sum);
    } else if (arg > 0 && arg <= limit) {
        return makeIntValue((p.*func)(arg));
    } else {
        return 0;
    }
}

IntValue*
IntBaseArrayProperty::get(IntDataSegment& dseg, uint32_t index, uint32_t nargs)
{
    int32_t arg;
    checkArgumentCount(nargs, 1);
    if (!checkIntArg(arg, dseg[index]))
        return 0;

    switch (ibp) {
     case ibpEngineStorage:
        return performArrayReference(&GPlanet::getBaseEngineStore, NUM_ENGINES, arg);
     case ibpHullStorage:
        return performArrayReference(&GPlanet::getBaseHullStoreSlot, NUM_HULLS_PER_PLAYER, arg);
     case ibpBeamStorage:
        return performArrayReference(&GPlanet::getBaseBeamStore, NUM_BEAMS, arg);
     case ibpLauncherStorage:
        return performArrayReference(&GPlanet::getBaseLauncherStore, NUM_TORPS, arg);
     case ibpAmmoStorage:
        return performArrayReference(&GPlanet::getBaseAmmoStore, NUM_TORPS+1, arg);
     default:
        return 0;
    }
}

void
IntBaseArrayProperty::set(IntDataSegment& /*dseg*/, uint32_t /*index*/, uint32_t /*nargs*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntContext*
IntBaseArrayProperty::makeFirstContext()
{
    throw IntError::typeError(IntError::tyIterable);
}

IntBaseArrayProperty*
IntBaseArrayProperty::clone()
{
    return new IntBaseArrayProperty(p, ibp);
}

string_t
IntBaseArrayProperty::toString(bool /*readable*/)
{
    return "#<array>";
}

void
IntBaseArrayProperty::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}


/** Get starbase property.
    \param pl Starbase
    \param ibp Starbase property to query */
IntValue*
getBaseProperty(const GPlanet& pl, IntBaseProperty ibp)
{
    if (!pl.isExistant() || !pl.hasBase())
        return 0;

    switch (ibp) {
     case ibpBaseDamage:
        return makeIntValue(pl.getBaseDamage());
     case ibpBaseDefense:
        return makeIntValue(pl.getStructures(str_BaseDefense));
     case ibpBaseDefenseMax:
        return makeIntValue(pl.getMaxStructures(str_BaseDefense));
     case ibpBaseFighters:
        return makeIntValue(pl.getCargo(el_Fighters));
     case ibpBaseFightersMax:
        return makeIntValue(config.MaximumFightersOnBase(pl.getOwner()));
     case ibpBeamTech:
        return makeIntValue(pl.getBaseTech(tl_Beam));
     case ibpBuildBeam:
        return makeIntValue(pl.getBaseBuildOrder().beam_type);
     case ibpBuildBeamCount:
        return makeIntValue(pl.getBaseBuildOrder().beam_count);
     case ibpBuildEngine:
        return makeIntValue(pl.getBaseBuildOrder().engine);
     case ibpBuildFlag:
        return makeBoolValue(pl.getBaseBuildHull() != 0);
     case ibpBuildHull:
        return makeIntValue(pl.getBaseBuildHull());
     case ibpBuildHullName:
        if (int n = pl.getBaseBuildHull())
            return makeStringValue(::getHull(n).getName());
        else
            return 0;
     case ibpBuildHullShort:
        if (int n = pl.getBaseBuildHull())
            return makeStringValue(::getHull(n).getShortName());
        else
            return 0;
     case ibpBuildQueuePos:
        /* FIXME */
        return 0;
     case ibpBuildTorp:
        return makeIntValue(pl.getBaseBuildOrder().torp_type);
     case ibpBuildTorpCount:
        return makeIntValue(pl.getBaseBuildOrder().torp_count);
     case ibpEngineTech:
        return makeIntValue(pl.getBaseTech(tl_Engine));
     case ibpHullTech:
        return makeIntValue(pl.getBaseTech(tl_Hull));
     case ibpMission:
        return makeIntValue(pl.getBaseMission());
     case ibpMissionName:
        return makeStringValue(getBaseMissionName(pl.getBaseMission()));
     case ibpShipyardAction:
        switch (pl.getBaseShipyardOrder()) {
         case sya_Fix:     return makeStringValue("Fix");
         case sya_Recycle: return makeStringValue("Recycle");
        }
        return 0;
     case ibpShipyardId:
        return makeIntValue(pl.getBaseShipyardId());
     case ibpShipyardName:
        if (int sid = pl.getBaseShipyardId())
            return makeStringValue(getDisplayedTurn().getCurrentUniverse().getShip(sid).getName());
        else
            return 0;
     case ibpShipyardStr:
        switch (pl.getBaseShipyardOrder()) {
         case sya_Fix:     return makeStringValue(format("Fix %s", getDisplayedTurn().getCurrentUniverse().getShip(pl.getBaseShipyardId()).getName()));
         case sya_Recycle: return makeStringValue(format("Recycle %s", getDisplayedTurn().getCurrentUniverse().getShip(pl.getBaseShipyardId()).getName()));
        }
        return 0;
     case ibpTorpedoTech:
        return makeIntValue(pl.getBaseTech(tl_Torpedo));

     case ibpEngineStorage:
     case ibpHullStorage:
     case ibpBeamStorage:
     case ibpLauncherStorage:
     case ibpAmmoStorage:
        return new IntBaseArrayProperty(pl, ibp);
    }
    return 0;
}
