/**
  *  \file int/if/hullif.cc
  *  \brief Interpreter Game Interface: Hull Properties
  */

#include "int/if/hullif.h"
#include "game/spec.h"
#include "int/interror.h"
#include "cpluslib/misc.h"
#include "int/data.h"
#include "int/compound.h"
#include "int/if/ifutil.h"
#include "int/if/specif.h"

/** Implementation of "Hull(nn)".
    \param dseg Data segment containing function parameters
    \param index Index of function parameters
    \param nargs Number of function parameters
    \return value representing "Hull(nn)" */
IntValue*
IFHullGet(IntDataSegment& dseg, uint32_t index, uint32_t nargs)
{
    int32_t id;

    checkArgumentCount(nargs, 1);
    if (!checkIntArg(id, dseg[index], 1, getNumHulls()))
        return 0;

    return IntSimpleContextValue::createFromNewContext(new IntHullContext(id));
}

/** Implementation of "ForEach Hull".
    \return Context representing first hull */
IntContext*
IFHullMake()
{
    if (getNumHulls() > 0)
        return new IntHullContext(1);
    else
        return 0;
}

/***************************** IntHullContext ****************************/

namespace {
    enum { HullProperty, ComponentProperty };
}

// FIXME: PCC 1.x items currently not implemented here:
//   Hull (same as Name)
//   Hull$ (same as Id)
//   Hull.Short (same as Name.Short)
//   Hull.Special (same as Special)
//   Hull.Special.Str (same as Special.Str)
static const TIntMapping hull_mapping[] = {
    { "BEAM.MAX",         ihpMaxBeams,         HullProperty },
    { "CARGO.MAX",        ihpMaxCargo,         HullProperty },
    { "CARGO.MAXFUEL",    ihpMaxFuel,          HullProperty },
    { "COST.D",           icpCostD,            ComponentProperty },
    { "COST.M",           icpCostM,            ComponentProperty },
    { "COST.MC",          icpCostMC,           ComponentProperty },
    { "COST.STR",         icpCostStr,          ComponentProperty },
    { "COST.T",           icpCostT,            ComponentProperty },
    { "CREW.NORMAL",      ihpMaxCrew,          HullProperty },
    { "ENGINE.COUNT",     ihpNumEngines,       HullProperty },
    { "FIGHTER.BAYS",     ihpNumFighterBays,   HullProperty },
    { "ID",               icpId,               ComponentProperty },
    { "IMAGE",            ihpImage,            HullProperty },
    { "IMAGE$",           ihpImage2,           HullProperty },
    { "MASS",             icpMass,             ComponentProperty },
    { "NAME",             icpName,             ComponentProperty },
    { "NAME.SHORT",       icpNameShort,        ComponentProperty },
    { "SPECIAL",          ihpSpecial,          HullProperty },
    { "TECH",             icpTech,             ComponentProperty },
    { "TECH.HULL",        icpTech,             ComponentProperty },
    { "TORP.LMAX",        ihpMaxTorpLaunchers, HullProperty },
};

/** Constructor.
    \param hid [1,getNumHulls()] Hull Id */
IntHullContext::IntHullContext(int hid)
    : hid(hid)
{ }

IntHullContext::~IntHullContext()
{ }

int32_t
IntHullContext::lookup(string_t name)
{
    return lookupName(name, hull_mapping, countof(hull_mapping));
}

void
IntHullContext::set(int32_t index, IntValue* value)
{
    switch (hull_mapping[index].domain) {
     case HullProperty:
        setHullProperty(getHull(hid), IntHullProperty(hull_mapping[index].index), value);
        break;
     default:
        throw IntError::notAssignable();
    }
}

IntValue*
IntHullContext::get(int32_t index)
{
    switch (hull_mapping[index].domain) {
     case HullProperty:
        return getHullProperty(getHull(hid), IntHullProperty(hull_mapping[index].index));
     case ComponentProperty:
        return getComponentProperty(getHull(hid), IntComponentProperty(hull_mapping[index].index));
     default:
        return 0;
    }
}

bool
IntHullContext::next()
{
    if (hid < getNumHulls()) {
        ++hid;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntHullContext::clone()
{
    return new IntHullContext(hid);
}

/** Get hull property.
    \param h Hull to query
    \param isp Hull property to query */
IntValue*
getHullProperty(const GHull& h, IntHullProperty isp)
{
    switch (isp) {
     case ihpMaxBeams:
        /*+ Maximum number of beams on this ship. */
        return makeIntValue(h.getMaxBeams());
     case ihpMaxCargo:
        /*+ Maximum cargo on this ship. */
        return makeIntValue(h.getCargo());
     case ihpMaxFuel:
        /*+ Maximum fuel on this ship. */
        return makeIntValue(h.getFuel());
     case ihpMaxCrew:
        /*+ Maximum crew on this ship. */
        return makeIntValue(h.getCrew());
     case ihpNumEngines:
        /*+ Number of engines. */
        return makeIntValue(h.getNumEngines());
     case ihpSpecial:
        /*+ String identifying the major special functions of this hull.
            The string will contain each letter if and only if the hull
            has the respective ability assigned for all players.
            - "C" (Cloak, including Advanced and Hardened Cloak)
            - "H" (Hyperdrive)
            - "G" (Gravitonic accelerator)
            - "B" (Bioscan, including Full Bioscan)
            - "A" (Alchemy, including Neutronic/Aries Refinery) */
        {
            string_t result;
            const GExpLevelSet levels = GExpLevelSet::allUpTo(NUM_EXPERIENCE_LEVELS);
            const GPlayerSet  players = GPlayerSet::allUpTo(NUM_PLAYERS) - 0;
            if ((h.getPlayersThatCan(hf_Cloak, true, levels)
                 + h.getPlayersThatCan(hf_AdvancedCloak, true, levels)
                 + h.getPlayersThatCan(hf_HardenedCloak, true, levels)).contains(players))
                result += "C";
            if (h.getPlayersThatCan(hf_Hyperdrive, true, levels).contains(players))
                result += "H";
            if (h.getPlayersThatCan(hf_Gravitonic, true, levels).contains(players))
                result += "G";
            if ((h.getPlayersThatCan(hf_Bioscan, true, levels)
                 + h.getPlayersThatCan(hf_FullBioscan, true, levels)).contains(players))
                result += "B";
            if ((h.getPlayersThatCan(hf_MerlinAlchemy, true, levels)
                 + h.getPlayersThatCan(hf_AriesRefinery, true, levels)
                 + h.getPlayersThatCan(hf_NeutronicRefinery, true, levels)).contains(players))
                result += "A";
            return makeStringValue(result);
        }
     case ihpMaxTorpLaunchers:
        /*+ Maximum number of torpedo launchers on this ship. */
        return makeIntValue(h.getMaxLaunchers());
     case ihpNumFighterBays:
        /*+ Number of fighter bays on this ship. */
        return makeIntValue(h.getNumBays());
     case ihpImage:
     case ihpImage2:
        // FIXME
        return makeIntValue(h.getPicture());
    }
    return 0;
}

/** Set hull property.
    \param h Hull to modify
    \param isp Property to modify
    \param value Value to assign */
void
setHullProperty(GHull& h, IntHullProperty isp, IntValue* value)
{
    (void) h;
    (void) isp;
    (void) value;
    throw IntError::notAssignable();
}
