/**
  *  \file int/if/ionif.cc
  *  \brief Interpreter Game Interface: Ion Storms
  */

#include "int/if/ionif.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/ion.h"
#include "game/tables.h"
#include "int/compound.h"
#include "int/if/ifutil.h"
#include "int/interror.h"
#include "int/value.h"

/** Implementation of "Storm(nn)".
    \param dseg Data segment containing function parameters
    \param index Index of function parameters
    \param nargs Number of function parameters
    \return value representing "Storm(nn)" */
IntValue*
IFIonGet(IntDataSegment& dseg, uint32_t index, uint32_t nargs)
{
    int32_t iid;
    checkArgumentCount(nargs, 1);
    if (!checkIntArg(iid, dseg[index], 1, NUM_STORMS))
        return 0;

    return IntSimpleContextValue::createFromNewContext(new IntIonContext(iid));
}

/** Implementation of "ForEach Storm".
    \return Context representing first ion storm */
IntContext*
IFIonMake()
{
    int iid = global_ion_type.getNextNoWrap(0, false);
    if (iid != 0)
        return new IntIonContext(iid);
    else
        return 0;
}

/***************************** IntIonContext *****************************/

namespace {
    enum { IonProperty };
}

static const TIntMapping ion_mapping[] = {
    { "CLASS",     iipClass,       IonProperty },
    { "HEADING",   iipHeadingName, IonProperty },
    { "HEADING$",  iipHeadingInt,  IonProperty },
    { "ID",        iipId,          IonProperty },
    { "LOC.X",     iipLocX,        IonProperty },
    { "LOC.Y",     iipLocY,        IonProperty },
    { "NAME",      iipName,        IonProperty },
    { "RADIUS",    iipRadius,      IonProperty },
    { "SPEED",     iipSpeedName,   IonProperty },
    { "SPEED$",    iipSpeedInt,    IonProperty },
    { "STATUS",    iipStatusName,  IonProperty },
    { "STATUS$",   iipStatusFlag,  IonProperty },
    { "VOLTAGE",   iipVoltage,     IonProperty },
};

/** Constructor.
    \param iid Ion Storm Id */
IntIonContext::IntIonContext(int iid)
    : iid(iid)
{ }

/** Destructor. */
IntIonContext::~IntIonContext()
{ }

int32_t
IntIonContext::lookup(string_t name)
{
    return lookupName(name, ion_mapping, countof(ion_mapping));
}

void
IntIonContext::set(int32_t index, IntValue* value)
{
    setIonProperty(global_ion_type.getObject(iid), IntIonProperty(ion_mapping[index].index), value);
}

IntValue*
IntIonContext::get(int32_t index)
{
    return getIonProperty(global_ion_type.getObject(iid), IntIonProperty(ion_mapping[index].index));
}

bool
IntIonContext::next()
{
    int n = global_ion_type.getNextNoWrap(iid, false);
    if (n != 0) {
        iid = n;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntIonContext::clone()
{
    return new IntIonContext(iid);
}

/** Get Ion Storm property.
    \param ion Storm to query
    \param iip Property to query
    \return property value */
IntValue*
getIonProperty(const GIonStorm& ion, IntIonProperty iip)
{
    if (!ion.isExistant())
        return 0;
    switch (iip) {
     case iipClass:
        if (ion.getVoltage() > 200)
            return makeIntValue(5);
        else
            return makeIntValue(ion.getVoltage() / 50 + 1);
     case iipHeadingInt:
        return makeIntValue(ion.getHeading());
     case iipHeadingName:
        return makeStringValue(getHeadingName(ion.getHeading()));
     case iipId:
        return makeIntValue(ion.getId());
     case iipLocX:
        return makeIntValue(ion.getPos().x);
     case iipLocY:
        return makeIntValue(ion.getPos().y);
     case iipName:
        return makeStringValue(ion.getName());
     case iipRadius:
        return makeIntValue(ion.getRadius());
     case iipSpeedInt:
        return makeIntValue(ion.getSpeed());
     case iipSpeedName:
        return makeStringValue(format("Warp %d", ion.getSpeed()));
     case iipStatusFlag:
        return makeBoolValue(ion.isGrowing());
     case iipStatusName:
        return makeStringValue(ion.isGrowing() ? "Growing" : "Weakening");
     case iipVoltage:
        return makeIntValue(ion.getVoltage());
    }
    return 0;
}

/** Set Ion Storm property. */
void
setIonProperty(GIonStorm& /*ion*/, IntIonProperty /*iip*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}
