/**
  *  \file int/if/planetif.cc
  *  \brief Interpreter Game Interface: Planets
  */

#include "int/if/planetif.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/planet.h"
#include "int/compound.h"
#include "int/data.h"
#include "int/if/baseif.h"
#include "int/if/ifutil.h"
#include "int/if/playerif.h"
#include "int/interror.h"
#include "int/intglobal.h"
#include "int/value.h"
#include "game/tables.h"
#include "game/ship.h"
#include "game/team.h"
#include "cpluslib/format.h"

/** Implementation of "Planet(nn)".
    \param dseg Data segment containing function parameters
    \param index Index of function parameters
    \param nargs Number of function parameters
    \return value representing "Planet(nn)" */
IntValue*
IFPlanetGet(IntDataSegment& dseg, uint32_t index, uint32_t nargs)
{
    int32_t id;
    checkArgumentCount(nargs, 1);
    if (!checkIntArg(id, dseg[index], 1, NUM_PLANETS))
        return 0;

    return IntSimpleContextValue::createFromNewContext(new IntPlanetContext(id));
}

/** Implementation of "ForEach Planet".
    \return Context representing first planet */
IntContext*
IFPlanetMake()
{
    int id = global_anyplanet_type.getNextNoWrap(0, false);
    if (id == 0)
        return 0;
    else
        return new IntPlanetContext(id);
}

/**************************** IntPlanetContext ***************************/

namespace { enum { PlanetProperty, BaseProperty, OwnerProperty }; }

static const TIntMapping planet_mapping[] = {
    { "BASE",                      ippBaseStr,           PlanetProperty },
    { "BASE.YESNO",                ippBaseFlag,          PlanetProperty },
    { "BUILD",                     ibpBuildHullName,     BaseProperty },
    { "BUILD.BEAM$",               ibpBuildBeam,         BaseProperty },
    { "BUILD.BEAM.COUNT",          ibpBuildBeamCount,    BaseProperty },
    { "BUILD.ENGINE$",             ibpBuildEngine,       BaseProperty },
    { "BUILD.HULL$",               ibpBuildHull,         BaseProperty },
    { "BUILD.QPOS",                ibpBuildQueuePos,     BaseProperty },
    { "BUILD.SHORT",               ibpBuildHullShort,    BaseProperty },
    { "BUILD.TORP$",               ibpBuildTorp,         BaseProperty },
    { "BUILD.TORP.COUNT",          ibpBuildTorpCount,    BaseProperty },
    { "BUILD.YESNO",               ibpBuildFlag,         BaseProperty },
    { "COLONISTS",                 ippColonists,         PlanetProperty },
    { "COLONISTS.CHANGE",          ippColonistChangeStr, PlanetProperty },
    { "COLONISTS.CHANGE$",         ippColonistChange,    PlanetProperty },
    { "COLONISTS.HAPPY",           ippColonistHappyStr,  PlanetProperty },
    { "COLONISTS.HAPPY$",          ippColonistHappy,     PlanetProperty },
    { "COLONISTS.SUPPORTED",       ippColonistSupported, PlanetProperty },
    { "COLONISTS.TAX",             ippColonistTax,       PlanetProperty },
    { "DAMAGE",                    ibpBaseDamage,        BaseProperty },
    { "DEFENSE",                   ippDefense,           PlanetProperty },
    { "DEFENSE.BASE",              ibpBaseDefense,       BaseProperty },
    { "DEFENSE.BASE.MAX",          ibpBaseDefenseMax,    BaseProperty },
    { "DEFENSE.BASE.WANT",         ippBaseDefenseWanted, PlanetProperty },
    { "DEFENSE.MAX",               ippDefenseMax,        PlanetProperty },
    { "DEFENSE.WANT",              ippDefenseWanted,     PlanetProperty },
    { "DENSITY.D",                 ippDensityD,          PlanetProperty },
    { "DENSITY.M",                 ippDensityM,          PlanetProperty },
    { "DENSITY.N",                 ippDensityN,          PlanetProperty },
    { "DENSITY.T",                 ippDensityT,          PlanetProperty },
    { "FACTORIES",                 ippFactories,         PlanetProperty },
    { "FACTORIES.MAX",             ippFactoriesMax,      PlanetProperty },
    { "FACTORIES.WANT",            ippFactoriesWanted,   PlanetProperty },
    { "FCODE",                     ippFCode,             PlanetProperty },
    { "FIGHTERS",                  ibpBaseFighters,      BaseProperty },
    { "FIGHTERS.MAX",              ibpBaseFightersMax,   BaseProperty },
    { "GROUND.D",                  ippGroundD,           PlanetProperty },
    { "GROUND.M",                  ippGroundM,           PlanetProperty },
    { "GROUND.N",                  ippGroundN,           PlanetProperty },
    { "GROUND.T",                  ippGroundT,           PlanetProperty },
    { "ID",                        ippId,                PlanetProperty },
    { "INDUSTRY",                  ippIndustry,          PlanetProperty },
    { "INDUSTRY$",                 ippIndustryCode,      PlanetProperty },
    { "LEVEL",                     ippLevel,             PlanetProperty },
    { "LOC.X",                     ippLocX,              PlanetProperty },
    { "LOC.Y",                     ippLocY,              PlanetProperty },
    { "MARKED",                    ippMarked,            PlanetProperty },
    { "MINED.D",                   ippMinedD,            PlanetProperty },
    { "MINED.M",                   ippMinedM,            PlanetProperty },
    { "MINED.N",                   ippMinedN,            PlanetProperty },
    { "MINED.STR",                 ippMinedStr,          PlanetProperty },
    { "MINED.T",                   ippMinedT,            PlanetProperty },
    { "MINES",                     ippMines,             PlanetProperty },
    { "MINES.MAX",                 ippMinesMax,          PlanetProperty },
    { "MINES.WANT",                ippMinesWanted,       PlanetProperty },
    { "MISSION",                   ibpMissionName,       BaseProperty },
    { "MISSION$",                  ibpMission,           BaseProperty },
    { "MONEY",                     ippMoney,             PlanetProperty },
    { "NAME",                      ippName,              PlanetProperty },
    { "NATIVES",                   ippNatives,           PlanetProperty },
    { "NATIVES.CHANGE",            ippNativeChangeStr,   PlanetProperty },
    { "NATIVES.CHANGE$",           ippNativeChange,      PlanetProperty },
    { "NATIVES.GOV",               ippNativeGov,         PlanetProperty },
    { "NATIVES.GOV$",              ippNativeGovCode,     PlanetProperty },
    { "NATIVES.HAPPY",             ippNativeHappyStr,    PlanetProperty },
    { "NATIVES.HAPPY$",            ippNativeHappy,       PlanetProperty },
    { "NATIVES.RACE",              ippNativeRace,        PlanetProperty },
    { "NATIVES.RACE$",             ippNativeRaceCode,    PlanetProperty },
    { "NATIVES.TAX",               ippNativeTax,         PlanetProperty },
    { "ORBIT",                     ippOrbitingShips,     PlanetProperty },
    { "ORBIT.ENEMY",               ippOrbitingEnemies,   PlanetProperty },
    { "ORBIT.OWN",                 ippOrbitingOwn,       PlanetProperty },
    { "OWNER",                     iplShortName,         OwnerProperty },
    { "OWNER$",                    iplId,                OwnerProperty },
    { "OWNER.ADJ",                 iplAdjName,           OwnerProperty },
    { "SHIPYARD",                  ibpShipyardStr,       BaseProperty },
    { "SHIPYARD.ACTION",           ibpShipyardAction,    BaseProperty },
    { "SHIPYARD.ID",               ibpShipyardId,        BaseProperty },
    { "SHIPYARD.NAME",             ibpShipyardName,      BaseProperty },
    { "STORAGE.AMMO",              ibpAmmoStorage,       BaseProperty },
    { "STORAGE.BEAMS",             ibpBeamStorage,       BaseProperty },
    { "STORAGE.ENGINES",           ibpEngineStorage,     BaseProperty },
    { "STORAGE.HULLS",             ibpHullStorage,       BaseProperty },
    { "STORAGE.LAUNCHERS",         ibpLauncherStorage,   BaseProperty },
    { "SUPPLIES",                  ippSupplies,          PlanetProperty },
    { "TASK",                      ippTask,              PlanetProperty },
    { "TASK.BASE",                 ippTaskBase,          PlanetProperty },
    { "TECH.BEAM",                 ibpBeamTech,          BaseProperty },
    { "TECH.ENGINE",               ibpEngineTech,        BaseProperty },
    { "TECH.HULL",                 ibpHullTech,          BaseProperty },
    { "TECH.TORPEDO",              ibpTorpedoTech,       BaseProperty },
    { "TEMP",                      ippTempStr,           PlanetProperty },
    { "TEMP$",                     ippTemp,              PlanetProperty },
};


IntPlanetContext::IntPlanetContext(int pid)
    : pid(pid)
{ }

IntPlanetContext::~IntPlanetContext()
{ }

int32_t
IntPlanetContext::lookup(string_t name)
{
    static const int32_t nproperties = countof(planet_mapping);

    /* Check prefix */
    if (name.compare(0, 7, "PLANET.", 7) == 0)
        name.erase(0, 7);

    /* Check user-defined properties */
    IntVariableNames::index_t ix = planet_property_names.lookup(name);
    if (ix != planet_property_names.nil)
        return ix + nproperties;

    /* Check predefined properties */
    return lookupName(name, planet_mapping, nproperties);
}

void
IntPlanetContext::set(int32_t index, IntValue* value)
{
    static const int32_t nproperties = countof(planet_mapping);
    if (index < nproperties) {
        /* Builtin property */
        // GPlanet& pl = getDisplayedTurn().getCurrentUniverse().getPlanet(pid);
        // FIXME
        throw IntError::notAssignable();
    } else {
        /* User property */
        getPlanetProperties(pid).set(index - nproperties, value);
    }
}

IntValue*
IntPlanetContext::get(int32_t index)
{
    static const int32_t nproperties = countof(planet_mapping);
    if (index < nproperties) {
        /* Builtin property */
        GPlanet& pl = getDisplayedTurn().getCurrentUniverse().getPlanet(pid);
        switch (planet_mapping[index].domain) {
         case PlanetProperty:
            return getPlanetProperty(pl, IntPlanetProperty(planet_mapping[index].index));
         case BaseProperty:
            return getBaseProperty(pl, IntBaseProperty(planet_mapping[index].index));
         case OwnerProperty:
            if (pl.isExistant())
                return getPlayerProperty(pl.getOwner(), IntPlayerProperty(planet_mapping[index].index));
            else
                return 0;
        }
        return 0;
    } else {
        /* User property */
        return IntValue::cloneOf(getPlanetProperties(pid)[index - nproperties]);
    }
}

bool
IntPlanetContext::next()
{
    int n = global_anyplanet_type.getNextNoWrap(pid, false);
    if (n != 0) {
        pid = n;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntPlanetContext::clone()
{
    return new IntPlanetContext(pid);
}


template<typename T>
static IntValue*
makeIntFromMaybe(T value)
{
    if (value.isKnown())
        return makeIntValue(value);
    else
        return 0;
}

static int
countShips(GPoint pt, GPlayerSet players)
{
    int n = 0;
    for (int sid = global_anyship_type.getNextNoWrap(0, false); sid != 0; sid = global_anyship_type.getNextNoWrap(sid, false)) {
        const GShip& sh = global_anyship_type.getObject(sid);
        if (pt == sh.getPos() && players.contains(sh.getOwner()))
            ++n;
    }
    return n;
}

IntValue*
getPlanetProperty(const GPlanet& pl, IntPlanetProperty ipp)
{
    TAutobuildSettings abs;

    /* For non-existant planets, we only know Name and Id */
    /* FIXME: is this needed? */
    if (!pl.isExistant() && ipp != ippId && ipp != ippName)
        return 0;

    switch (ipp) {
     case ippBaseDefenseWanted:
        pl.getAutobuild(abs);
        return makeIntValue(abs.goal[str_BaseDefense]);
     case ippBaseFlag:
        return makeBoolValue(pl.hasBase());
     case ippBaseStr:
        if (pl.hasBase())
            return makeStringValue("present");
        else if (pl.isBuildingBase())
            return makeStringValue("being built");
        else
            return makeStringValue("-");
     case ippColonistChange:
        return makeIntFromMaybe(pl.getColonistChange());
     case ippColonistChangeStr:
        {
            mn32_t change = pl.getColonistChange();
            if (change.isKnown())
                return makeStringValue(getHappyChangeName(change));
            else
                return 0;
        }
     case ippColonistHappy:
        return makeIntFromMaybe(pl.getColonistHappiness());
     case ippColonistHappyStr:
        if (pl.getColonistHappiness().isKnown())
            return makeStringValue(getHappinessName(pl.getColonistHappiness()));
        else
            return 0;
     case ippColonistSupported:
        return makeIntFromMaybe(pl.getMaxSupportedColonists());
     case ippColonistTax:
        return makeIntFromMaybe(pl.getColonistTax());
     case ippColonists:
        return makeIntFromMaybe(pl.getColonists());
     case ippDefense:
        return makeIntFromMaybe(pl.getStructures(str_Defense));
     case ippDefenseMax:
        return makeIntFromMaybe(pl.getMaxStructures(str_Defense));
     case ippDefenseWanted:
        pl.getAutobuild(abs);
        return makeIntValue(abs.goal[str_Defense]);
     case ippDensityD:
        return makeIntFromMaybe(pl.getOreDensity(el_Duranium));
     case ippDensityM:
        return makeIntFromMaybe(pl.getOreDensity(el_Molybdenum));
     case ippDensityN:
        return makeIntFromMaybe(pl.getOreDensity(el_Neutronium));
     case ippDensityT:
        return makeIntFromMaybe(pl.getOreDensity(el_Tritanium));
     case ippFCode:
        if (pl.isFCodeKnown())
            return makeStringValue(pl.getFCode());
        else
            return 0;
     case ippFactories:
        return makeIntFromMaybe(pl.getStructures(str_Factories));
     case ippFactoriesMax:
        return makeIntFromMaybe(pl.getMaxStructures(str_Factories));
     case ippFactoriesWanted:
        pl.getAutobuild(abs);
        return makeIntValue(abs.goal[str_Factories]);
     case ippGroundD:
        return makeIntFromMaybe(pl.getOreGround(el_Duranium));
     case ippGroundM:
        return makeIntFromMaybe(pl.getOreGround(el_Molybdenum));
     case ippGroundN:
        return makeIntFromMaybe(pl.getOreGround(el_Neutronium));
     case ippGroundT:
        return makeIntFromMaybe(pl.getOreGround(el_Tritanium));
     case ippId:
        return makeIntValue(pl.getId());
     case ippIndustry:
        if (pl.getIndustryLevel().isKnown())
            return makeStringValue(getIndustryLevelName(pl.getIndustryLevel()));
        else
            return 0;
     case ippIndustryCode:
        return makeIntFromMaybe(pl.getIndustryLevel());
     case ippLevel:
        {
            int16_t level = pl.unit_scores.getScore(planet_score_definitions.lookupScore(ScoreId_ExpLevel));
            if (level >= 0)
                return makeIntValue(level);
            else
                return 0;
        }
     case ippLocX:
        return makeIntValue(pl.getPos().x);
     case ippLocY:
        return makeIntValue(pl.getPos().y);
     case ippMarked:
        return makeBoolValue(pl.isMarked());
     case ippMinedD:
        if (pl.isCargoAmountKnown(el_Duranium))
            return makeIntValue(pl.getCargo(el_Duranium));
        else
            return 0;
     case ippMinedM:
        if (pl.isCargoAmountKnown(el_Molybdenum))
            return makeIntValue(pl.getCargo(el_Molybdenum));
        else
            return 0;
     case ippMinedN:
        if (pl.isCargoAmountKnown(el_Neutronium))
            return makeIntValue(pl.getCargo(el_Neutronium));
        else
            return 0;
     case ippMinedStr:
        {
            // FIXME: 
            bool haveAny = false;
            string_t result;
            for (GCargoType el = el_Neutronium; el <= el_Molybdenum; ++el) {
                if (pl.isCargoAmountKnown(el)) {
                    if (pl.getCargo(el) >= 0) {
                        if (result.size())
                            result += ' ';
                        result += "NTDM"[el];
                        result += format("%d", int32_t(pl.getCargo(el)));
                    }
                    haveAny = true;
                }
            }
            if (haveAny)
                return makeStringValue(result);
            else
                return 0;
        }
     case ippMinedT:
        if (pl.isCargoAmountKnown(el_Tritanium))
            return makeIntValue(pl.getCargo(el_Tritanium));
        else
            return 0;
     case ippMines:
        return makeIntFromMaybe(pl.getStructures(str_Mines));
     case ippMinesMax:
        return makeIntFromMaybe(pl.getMaxStructures(str_Mines));
     case ippMinesWanted:
        pl.getAutobuild(abs);
        return makeIntValue(abs.goal[str_Mines]);
     case ippMoney:
        if (pl.isCargoAmountKnown(el_Money))
            return makeIntValue(pl.getCargo(el_Money));
        else
            return 0;
     case ippName:
        return makeStringValue(pl.getName());
     case ippNativeChange:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0)
            return makeIntFromMaybe(pl.getNativeChange());
        else
            return 0;
     case ippNativeChangeStr:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0 && pl.getNativeChange().isKnown())
            return makeStringValue(getHappyChangeName(pl.getNativeChange()));
        else
            return 0;
     case ippNativeGov:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0 && pl.getNativeGovernment().isKnown())
            return makeStringValue(getNativeGovernmentName(pl.getNativeGovernment()));
        else
            return 0;
     case ippNativeGovCode:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0)
            return makeIntFromMaybe(pl.getNativeGovernment());
        else
            return 0;
     case ippNativeHappy:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0)
            return makeIntFromMaybe(pl.getNativeHappiness());
        else
            return 0;
     case ippNativeHappyStr:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0 && pl.getNativeHappiness().isKnown())
            return makeStringValue(getHappinessName(pl.getNativeHappiness()));
        else
            return 0;
     case ippNativeRace:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0 && pl.getNativeRace().isKnown())
            return makeStringValue(getNativeRaceName(pl.getNativeRace()));
        else
            return 0;
     case ippNativeRaceCode:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0)
            return makeIntFromMaybe(pl.getNativeRace());
        else
            return 0;
     case ippNativeTax:
        if (pl.getNatives().isKnown() && pl.getNatives() > 0)
            return makeIntFromMaybe(pl.getNativeTax());
        else
            return 0;
     case ippNatives:
        return makeIntFromMaybe(pl.getNatives());
     case ippOrbitingEnemies:
        return makeIntValue(countShips(pl.getPos(), GPlayerSet::allUpTo(NUM_OWNERS) - getPlayerId()));
     case ippOrbitingOwn:
        return makeIntValue(countShips(pl.getPos(), GPlayerSet(getPlayerId())));
     case ippOrbitingShips:
        return makeIntValue(countShips(pl.getPos(), GPlayerSet::allUpTo(NUM_OWNERS)));
     case ippSupplies:
        if (pl.isCargoAmountKnown(el_Supplies))
            return makeIntValue(pl.getCargo(el_Supplies));
        else
            return 0;
     case ippTask:
     case ippTaskBase:
        // FIXME
        return 0;
     case ippTemp:
        return makeIntFromMaybe(pl.getTemperature());
     case ippTempStr:
        if (pl.getTemperature().isKnown())
            return makeStringValue(getTemperatureName(pl.getTemperature()));
        else
            return 0;
    }
    return 0;
}
