/**
  *  \file int/intglobal.cc
  *  \brief Interpreter: Global Definitions
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/intglobal.h"
#include "int/names.h"
#include "int/data.h"
#include "game/struct.h"
#include "cpluslib/assert.h"
#include "cpluslib/misc.h"
#include "int/context.h"

/*************************** Exported variables **************************/

/** Names of all planet properties. */
IntVariableNames planet_property_names;

/** Names of all ship properties. */
IntVariableNames ship_property_names;

/** Names of all global variables. */
IntVariableNames global_property_names;

/** All global values. */
IntDataSegment global_values;

/** All globally-active contexts. This is a template for new IntExecutionContext's. */
ptr_vector<IntContext> global_contexts;


/************************** IntTemporaryContext **************************/

/** Constructor. Make \c con temporarily active. */
IntTemporaryContext::IntTemporaryContext(IntContext* con)
    : con(con)
{
    global_contexts.push_back(con);
}

/** Destructor. Deactivates (and destroys) the context. */
IntTemporaryContext::~IntTemporaryContext()
{
    ASSERT(global_contexts.back() == con);
    global_contexts.pop_back();
}


/*************************** Accessor Functions **************************/

/** Planet property values. */
static IntDataSegment planet_properties[NUM_PLANETS];

/** Ship property values. */
static IntDataSegment ship_properties[NUM_SHIPS];


/** Initialize interpreter globals. Call this once per program invocation. */
void
initInterpreter()
{
    ship_property_names.add("COMMENT");
    planet_property_names.add("COMMENT");
    for (char c = 'A'; c < 'Z'; ++c)
        global_property_names.add(string_t(1, c));
}

/** Discard interpreter's game-specific data. Call this after closing
    down the game. */
void
clearInterpreterGameData()
{
    // Clear all properties
    for (unsigned i = 0; i < countof(planet_properties); ++i)
        planet_properties[i].clear();
    for (unsigned i = 0; i < countof(ship_properties); ++i)
        ship_properties[i].clear();
}

/** Get property data segment for a ship. */
IntDataSegment&
getShipProperties(int sid)
{
    ASSERT(sid > 0 && sid <= NUM_SHIPS);
    return ship_properties[sid-1];
}

/** Get property data segment for a planet. */
IntDataSegment&
getPlanetProperties(int pid)
{
    ASSERT(pid > 0 && pid <= NUM_PLANETS);
    return planet_properties[pid-1];
}
