/**
  *  \file int/opcode.h
  *  \brief Interpreter: Opcodes
  *
  *  This defines the data structure used to represent individual
  *  instructions.
  */
#ifndef PCC_V2_INT_OPCODE_H
#define PCC_V2_INT_OPCODE_H

#include <iosfwd>
#include "cpluslib/types.h"

/* Avoid gcc/Linux namespace pollution */
#undef major
#undef minor

/** Opcode for compiled CCScript. Each instruction is represented by
    one object of this type, totalling to 32 bit per instruction. */
struct IntOpcode {
    uint8_t  major;                ///< Major opcode.
    uint8_t  minor;                ///< Minor opcode/parameter.
    uint16_t arg;                  ///< Parameter.

    /** Major opcodes. */
    enum Major {
        maPush,                    ///< Push.              minor=Scope,          arg=table index/value.
        maBinary,                  ///< Binary operation.  minor=IntBinaryOperation.
        maUnary,                   ///< Unary operation.   minor=IntUnaryOperation.
        maJump,                    ///< Jump.              minor=jump flags,     arg=target address/name.
        maCallProc,                ///< Call procedure.    minor=number of args, arg=name.
        maCallFunc,                ///< Call function.     minor=number of args, arg=name.
        maIndirect,                ///< Indirect call.     minor=Indirect,       arg=number of args.
        maStack,                   ///< Stack operations.  minor=Stack,          arg=parameter.
        maPop,                     ///< Pop.               minor=Scope,          arg=table index.
        maStore,                   ///< Store.             minor=Scope,          arg=name table index.
        maMemref,                  ///< Member references. minor=Indirect,       arg=member name table index.
        maDim,                     ///< Make variable.     minor=Scope,          arg=name.
        maSpecial                  ///< Specials.          minor=Special.
    };

    /** Scope. Used as minor opcode for Push, Pop, Store, Dim. This defines the interpretation of arg. */
    enum Scope {
        sNamedVariable,            ///< Named variable. Parameter is index into BCO's name list. Used for Push/Pop/Store.
        sLocal,                    ///< Local variable. Parameter is index into local frame data. Used for Push/Pop/Store/Dim.
        sStatic,                   ///< Static variable. Parameter is index into static (topmost frame) data. Used for Push/Pop/Store/Dim.
        sShared,                   ///< Shared variable. Parameter is index into shared data. Used for Push/Pop/Store/Dim.
        sLiteral,                  ///< Literal. Parameter is index into BCO's literal list. Used for Push.
        sInteger,                  ///< Integer. Parameter is literal. Used for Push.
        sBoolean,                  ///< Boolean. Parameter is literal. Used for Push.
        sEmpty,                    ///< Empty. No parameter. Used for Push.
        sSubroutine                ///< Subroutine. Parameter is index into BCO's subroutine list. Used for Push.
    };

    /** Jump flags. Used as minor opcode for Jump. */
    enum {
        // Regular jumps:
        jIfTrue    = 1,            ///< Jump if true.
        jIfFalse   = 2,            ///< Jump if false.
        jIfEmpty   = 4,            ///< Jump if empty.
        jAlways    = 7,            ///< Jump always (sum of the above).
        jPopAlways = 8,            ///< Pop after checking condition.

        // Special jumps:
        jCatch     = 16,           ///< Push EH frame, jump on throw.
        jLabel     = 0,            ///< This is a label (jump never).

        // General flag:
        jSymbolic  = 128           ///< If set, address is not relocated yet.
    };

    /** Stack operations. Used as minor opcode for Stack. */
    enum Stack {
        miStackDup,                ///< Duplicate nn'th element.
        miStackDrop,               ///< Drop nn elements.
        miStackSwap                ///< Swap nn'th element and TOS.
    };

    /** Indirect/Member operations. Used as minor opcode for Indirect/Memref.
        For Indirect, arg is number of parameters on stack. For Memref, arg
        is name table index of member name. Stack contains args (for Indirect),
        then object to call/dereference, then value to assign (for store/pop). */
    enum Indirect {
        miIMCall,                  ///< Call fun(args), no result.
        miIMLoad,                  ///< Call fun(args), one result.
        miIMStore,                 ///< Assign fun(args) := value, keep value.
        miIMPop                    ///< Assign fun(args) := value, drop value.
    };

    /** Specials. */
    enum Special {
        miSpecialUncatch,          ///< Cancel previous catch.
        miSpecialReturn,           ///< Terminate current frame.
        miSpecialWith,             ///< Turn TOS into context.
        miSpecialEndWith,          ///< Cancel miSpecialWith.
        miSpecialFirstIndex,       ///< Iterate over TOS.
        miSpecialNextIndex,        ///< Next iteration.
        miSpecialEndIndex          ///< Cancel miSpecialFirstIndex prematurely.
    };
};

std::ostream& operator<<(std::ostream& out, const IntOpcode& opcode);

#endif
